# -*- coding: utf-8 -*-
"""
Created on Tue Apr 24 15:40:43 2012

@author: Edward
"""

#Final Project for ChE210D
"""
The objective of this simulation is to measure the force between two plates that
are each functionalized with a polymer brush. This system is studied using a
molecular dynamics simulation. The polymer brush is tethered to the wall, but 
is allowed to move in the plane of the wall. 
"""
#import python modules
import numpy as np
import time

#import compiled fortran library
import ex5lib

import sys
import os
import atomwrite
#whether or not to use the 3d visualization
UseVisual = True
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
#set the random number seed; useful for debugging 342324
np.random.seed()


#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.



def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global M, L, Cut, N, d
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [ex5lib.calcenergy(Pos + NormDir * x, M, L, L, Cut,3,N) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = ex5lib.calcenergy(Pos + NormDir * dx, M, L, L, Cut,3,N)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [ex5lib.calcenergy(Pos + NormDir * 2. * dx, M, L, L, Cut,3,N)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [ex5lib.calcenergy(Pos + NormDir * Dists[-1], M, L, L, Cut,3,N)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = ex5lib.calcenergy(Pos + NormDir * d3, M, L, L, Cut,3,N)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = ex5lib.calcenergy(Pos + NormDir * d3, M, L, L, Cut,3,N)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)
    PosMin = Pos + NormDir * Dists[1]
    i=0
    while i<N:
        PosMin[i,:]=Pos[i,:] #reset positions of brush ends to wall
        i+=M
    PEMin = PEs[1]
    
    #for the sake of visualization, switch z and y axes
    PosVis=np.zeros_like(PosMin)
    PosVis[:,0]=PosMin[:,0]
    PosVis[:,1]=PosMin[:,2]
    PosVis[:,2]=PosMin[:,1]
    
    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosVis)
        else:
            #initialize the visualization window
            atomvis.Init(PosVis)
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global M, L, Cut, N, d
    print "Welcome to Conjugate Gradient"
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = ex5lib.calcenergyforces(Pos, M, L, L, Cut, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    a=0
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = ex5lib.calcenergyforces(Pos, M, L, L, Cut, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
        PosVis=np.zeros_like(Pos)
        PosVis[:,0]=Pos[:,0]
        PosVis[:,1]=Pos[:,2]
        PosVis[:,2]=Pos[:,1]
        if a % 10 ==0:
            Pdb.write(PosVis)
        a+=1
    return PE, Pos

def InitPositions(N, M, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    M: number of monomers
    L: box length
    d: distance between plates
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for brushes at wall
    NLat = ((N/2/M)**(1./2.))
    LatSpac = L / NLat
    #make an array of lattice sites for brushes
    r = np.arange(LatSpac/2,L+LatSpac/2, LatSpac, dtype=float)
    #lattice spacing for the z-direction where d is the z distance b/w walls
    ZSpac = L/(2*M)
    #array of lattice sites in z
    rz = np.arange(0,L/2, ZSpac, dtype=float)
    rz2 = -1 * np.arange(-L,-L/2,ZSpac, dtype=float)
    rlist=np.append(rz,rz2)
    #list of z positions such that each index divisible by 5 is the brush end
    #That is: Pos[i] where i%5=0 is always at the wall (top or bottom)
    
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in rlist:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                if i%M>0:
                    Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >=N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T, dt, shearv):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    global N, M
    #recenter to zero net momentum in y and z directions (assuming all masses same)
    Vel[:,1] = Vel[:,1] - Vel.mean(axis=0)[1]
    Vel[:,2] = Vel[:,2] - Vel.mean(axis=0)[2]
    if shearv == 0:
        Vel[:,0] = Vel[:,0] - Vel.mean(axis=0)[0]
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    #VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    tau = 0.01
    if shearv == 0:
        VScale = np.sqrt(1. + dt/tau*((1.5 * (3*N-N/M) * T / KE)-1.))
    else: 
        VScale = np.sqrt(1. + dt/tau*((1.5 * (2*N-N/M) * T / KE)-1.))
    Vel[:,1] = Vel[:,1] * VScale
    Vel[:,2] = Vel[:,2] * VScale
    if shearv == 0:
        Vel[:,0] = Vel[:,0] * VScale
    i=0
    while i<N:
        Vel[i,:] = 0 #set velocity=0 for brush ends at wall
        i+=M
    return Vel


def InitVelocities(N, T, dt, shearv):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T, dt, shearv)
    i=0
    while i<N:
        Vel[i,:] = 0 #set velocity=0 for brush ends at wall
        i+=M
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global M, L, Cut,N, d
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = ex5lib.calcenergyforces(Pos, M, L, L, Cut, Forces)
    i=0
    while i<N:
        Accel[i,:] = 0 #set acceleration=0 for brush ends at wall
        i+=M
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))

#### YOUR CODE HERE ####

#### YOU WILL NEED TO DEFINE THE FOLLOWING GLOBAL ####
#### VARIABLES FOR THE CODE ABOVE TO WORK         ####

def Step1(N,M,L,Cut):
    """This is a function that intializes the system for the polymers in an 
    L^3 box, then rescales the positions to bring the  brushes into contact with
    one another.
    Input: 
        N: number of beads
        M: number of beads per polymer
        L: size of box
        Cut: the cutoff length of the LJ 12-6 potential
    Output:
        PE: Potential Energy after the brushes are in contact
        Pos: The new position array with the brushes touching
        d: the distance between the upper and lower walls
    """
    dx=0.001
    EFracTolLS=1.e-3
    EFracTolCG=1.e-5
    Pos = InitPositions(N,M,L)
    PE, Pos = ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG)
    print "Conjugate Gradient Complete"
    Dist=[]
    i=M-1
    while i<(N-M):
        dist=Pos[i+M,2]-Pos[i,2]
        Dist.append(dist)
        i+=2*M
    Min=np.min(Dist) - Cut
    #minimum distance between the upper and lower brushes - Cut
    #Now bring the upper and lower wall together. 
    a=M
    while a<N:
        Pos[a,2]=Pos[a,2] - Min 
        # move the upper brush down such that its doesnt overlap the bottom
        a+=1 #cycles through all the beads in a brush on the upper wall
        if a % (2*M) == 0: # cycles through each brush at the upper wall
            a+=M
    PosVis=np.zeros_like(Pos)
    PosVis[:,0]=Pos[:,0]
    PosVis[:,1]=Pos[:,2]
    PosVis[:,2]=Pos[:,1]
    atomvis.Update(PosVis)
    d=L-Min
    PE=ex5lib.calcenergy(Pos, M, L, d, Cut,3,N)
    return PE, Pos, d

def MoveWall(Pos, Forces, d):
    Dist=[]
    i=M-1
    while i<(N-M):
        dist=Pos[i+M,2]-Pos[i,2]
        Dist.append(dist)
        i+=2*M
    Min=np.min(Dist)
    #minimun distance between the upper and lower brushes
    #Now bring the upper and lower wall together. 
    Move=Min/2 #half the smallest distance between brushes
    a=M
    while a<N:
        Pos[a,2]=Pos[a,2] - Move 
        # move the upper brush down such that its doesnt overlap the bottom
        a+=1 #cycles through all the beads in a brush on the upper wall
        if a % (2*M) == 0: # cycles through each brush at the upper wall
            a+=M
    PosVis=np.zeros_like(Pos)
    PosVis[:,0]=Pos[:,0]
    PosVis[:,1]=Pos[:,2]
    PosVis[:,2]=Pos[:,1]
    atomvis.Update(PosVis)
    d=d-Move
    PE, Forces=ex5lib.calcenergyforces(Pos, M, L, d, Cut, Forces)
    return PE, Forces, Pos, d

def SetHeight(Pos, Forces, d, D):
    """
    Inputs
    Pos: Position Array
    d: current distance between walls
    D: the desired distance between walls
    """
    Move = d-D
    a=M
    while a<N:
        Pos[a,2]=Pos[a,2] - Move 
        # move the upper brush down
        a+=1 #cycles through all the beads in a brush on the upper wall
        if a % (2*M) == 0: # cycles through each brush at the upper wall
            a+=M
    PosVis=np.zeros_like(Pos)
    PosVis[:,0]=Pos[:,0]
    PosVis[:,1]=Pos[:,2]
    PosVis[:,2]=Pos[:,1]
    Pdb.write(PosVis)
    atomvis.Update(PosVis)
    PE, Forces=ex5lib.calcenergyforces(Pos, M, L, D, Cut, Forces)
    return PE, Forces, Pos, D

def Equil(N,M,L,D):
    Cut=2**(1./6.) #this cutoff distance is chosen to simulate a good solvent
    T=1.0
    dt=0.001
    shearv=0 #shear velocity applied to the upper wall in the x-direction
    RescaleFreq=1000
    TrajFreq=250
    MoveFreq=2000
    NStepsEquil1=50000
    NStepsEquil2=100000
    PE, Pos, d = Step1(N,M,L,Cut) #
    rho=N/(L*L*d)
    print PE
    Vel = InitVelocities(N,T,dt,shearv)
    Accel = InitAccel(Pos)
    n=0
    print "Begin Equilibration 1"
    while n<NStepsEquil1:
        Pos, Vel, Accel, KEnergy, PE = ex5lib.vvintegrate(Pos, Vel, Accel, M, L, d, Cut, dt)
        i=0
        while i<N:
            Vel[i,:] = 0 #set veex5lib.calcenergy(Pos + NormDir * d3, M, L, L, Cut,3,N)locity=0 for brush ends at wall
            Accel[i,:] = 0 #set acceleration=0
            i+=M
        n += 1
        Forces=Accel
        #for the sake of visualization, switch z and y axes
        PosVis=np.zeros_like(Pos)
        PosVis[:,0]=Pos[:,0]
        PosVis[:,1]=Pos[:,2]
        PosVis[:,2]=Pos[:,1]
        if UseVisual:
            atomvis.Update(PosVis)
        if n % TrajFreq ==0:
            Pdb.write(PosVis)
        if n % RescaleFreq == 0:
            PE, Forces = ex5lib.calcenergyforces(Pos, M, L, d, Cut, Forces)
            Vel = RescaleVelocities(Vel, T, dt, shearv)
        if n % MoveFreq == 0 and n<10000:
            PE, Forces, Pos, d = MoveWall(Pos, Forces, d)
        if n == 10000:
            print "Wall moves turned off"
            PE, Forces, Pos, D = SetHeight(Pos, Forces, d, D)
    n=0
    print PE
    print "Begin Equilibration 2"
    print "Distance between walls= %f" %(D)
    shearv=0.1
    print "Shear Velocity= %f" %(shearv)
    while n<NStepsEquil2:
        #Begin shearing the upper wall with a shear velocity equal to shearv
        a=M
        while a<N:
            Vel[a,0]=Vel[a,0] + shearv # add a shear velocity in x-direction 
            a+=2*M #cycles through all the beads tethered to the upper wall
        Pos, Vel, Accel, KEnergy, PE = ex5lib.vvintegrate(Pos, Vel, Accel, M, L, d, Cut, dt)
        i=0
        while i<N:
            Vel[i,:] = 0 #set velocity=0 for brush ends at wall
            Accel[i,:] = 0 #set acceleration=0
            i+=M
        n += 1
        #for the sake of visualization, switch z and y axes
        PosVis=np.zeros_like(Pos)
        PosVis[:,0]=Pos[:,0]
        PosVis[:,1]=Pos[:,2]
        PosVis[:,2]=Pos[:,1]
        # check for minimum image in x direction for sake of visualization
        b=M
        while b<N:
            if PosVis[b,0]>L:
                PosVis[b,0]=PosVis[b,0]-L
            b+=1
            if b % M == 0:
                b+=M
        if UseVisual:
            atomvis.Update(PosVis)
        if n % (RescaleFreq/5) == 0:
            PE, Forces = ex5lib.calcenergyforces(Pos, M, L, d, Cut, Forces)
            #Change sign of forces for one wall, then sum
            Forcesz=[]
            Forcesx=[]
            c=0
            while c<N:
                a=-1
                if c%(2*M) == 0:
                    a=1
                Forcesz.append(a*Forces[c,2])
                Forcesx.append(a*Forces[c,0])
                c+=M #cycles through all wall beads
            print "%i %f %f %f" %(n,PE, np.sum(Forcesz), np.sum(Forcesx))
        if n % TrajFreq ==0:
            Pdb.write(PosVis)
        if PE > 1e10:
            print "Potential energy has blown up"
            return 0
    Pdb.close()
    return Pos

N=160 #Note: N/(2*M) must be equal to a perfect square for code to work
M=5
L=(N/0.5)**(1./3.)
D=5
A=['C','C','C','C','C','O','O','O','O','O']
Names=A*(N/10)
Pdb=atomwrite.pdbfile("anim4.pdb",AtomNames=Names)
Equil (N,M,L,D)
print "end"